import random
from dataclasses import asdict
from typing import Any, Dict, Tuple, Union

import numpy as np
import torch as T
from gym import Env, spaces


def get_device(device: Union[T.device, str]) -> T.device:
    """
    Retrieve PyTorch device.
    It checks that the requested device is available first.
    For now, it supports only cpu and cuda.

    :param device: One for 'auto', 'cuda', 'cpu'
    :return:
    """
    if isinstance(device, T.device):
        return device

    # Cuda by default
    if device == "auto":
        device = "cuda"
    # Force conversion to th.device
    device = T.device(device)

    # Cuda not available
    if device.type == T.device("cuda").type and not T.cuda.is_available():
        return T.device("cpu")

    return device


def to_torch(*data, **kwargs) -> Union[Tuple[T.Tensor], T.Tensor]:
    """Convert to torch tensors"""
    device = get_device(kwargs.pop("device", "auto"))
    result = [None] * len(data)
    for i, el in enumerate(data):
        if isinstance(el, np.ndarray):
            result[i] = T.from_numpy(el).to(device, non_blocking=True)
        else:
            result[i] = T.as_tensor(el).to(device, non_blocking=True)

    if len(data) == 1:
        return result[0]
    else:
        return tuple(result)


def to_numpy(*data) -> Union[Tuple[np.ndarray], np.ndarray]:
    """Convert to numpy arrays"""
    result = [None] * len(data)
    for i, el in enumerate(data):
        if isinstance(el, T.Tensor):
            result[i] = el.detach().cpu().numpy()
        else:
            result[i] = np.array(el)

    if len(data) == 1:
        return result[0]
    else:
        return tuple(result)


def get_space_shape(
    space: spaces.Space,
) -> Tuple[int, ...]:
    """
    Get the shape of a space (useful for the buffers).
    :param space:
    :return:
    """
    if isinstance(space, spaces.Tuple):
        return (len(space),) + get_space_shape(space.spaces[0])
    if isinstance(space, spaces.Box):
        return space.shape
    elif isinstance(space, spaces.Discrete):
        # an int
        return (1,)
    elif isinstance(space, spaces.MultiDiscrete):
        # Number of discrete features
        return (int(len(space.nvec)),)
    elif isinstance(space, spaces.MultiBinary):
        # Number of binary features
        return (int(space.n),)
    elif isinstance(space, spaces.Dict):
        return get_space_shape(space["observation"])
    else:
        raise NotImplementedError(f"{space} observation space is not supported")


def get_space_range(space: spaces.Space) -> Tuple[Any, Any]:
    """
    Get the range of a space (useful for the buffers).
    :param space:
    :return:
    """
    if isinstance(space, spaces.Tuple):
        return get_space_range(space.spaces[0])
    if isinstance(space, spaces.Box):
        return space.low, space.high
    elif isinstance(space, spaces.Discrete):
        return 0, space.n - 1
    elif isinstance(space, spaces.MultiDiscrete):
        return np.zeros(space.nvec.shape), space.nvec - 1
    elif isinstance(space, spaces.MultiBinary):
        return 0, 1
    elif isinstance(space, spaces.Dict):
        return get_space_range(space["observation"])
    else:
        raise NotImplementedError(f"{space} observation space is not supported")


def extend_shape(original_shape: Tuple, new_size: int, axis: int = 0) -> Tuple:
    """Extend a dimension of a shape tuple"""

    shape = list(original_shape)
    shape[axis] = new_size
    return tuple(shape)


def filter_dataclass_by_none(class_object: Any) -> Dict[str, Any]:
    dict_class = asdict(class_object)
    return {k: v for k, v in dict_class.items() if v is not None}


def set_seed(seed: int, env: Env) -> None:
    """
    Set the seed for all the random generators.
    :param seed: The seed to set
    :param env: The environment to set the seed for
    """
    random.seed(seed)
    T.manual_seed(seed)
    T.cuda.manual_seed(seed)
    np.random.seed(seed)
    T.backends.cudnn.deterministic = True
    T.backends.cudnn.benchmark = False
    env.seed(seed)
    env.action_space.seed(seed)
    env.observation_space.seed(seed)


def filter_rewards(rewards: np.ndarray, dones: np.ndarray) -> np.ndarray:
    """
    Filter rewards based on done flags, all rewards after a done are set to 0.
    :param rewards: The rewards to filter (num_envs, num_steps) or (num_steps,)
    :param dones: The done flags (num_envs, num_steps) or (num_steps,)
    :return: The filtered rewards
    """
    if rewards.ndim == 1:
        rewards = rewards[np.newaxis, :]
        dones = dones[np.newaxis, :]
    for i, env_dones in enumerate(dones):
        done_indices = np.where(env_dones == 1)[0]
        if done_indices.size > 0:
            done_index = done_indices[0] + 1
            if done_index < rewards.shape[1]:
                rewards[i, done_index:] = 0

    return rewards.squeeze()



from typing import Dict, Tuple, Union

import numpy as np
import torch as T
from gym.core import GoalEnv

from pearll.buffers.base_buffer import BaseBuffer
from pearll.common.enumerations import GoalSelectionStrategy, TrajectoryType
from pearll.common.type_aliases import DictTrajectories, Tensor


class HERBuffer(BaseBuffer):
    """
    Hindsight Exprience Replay (HER) Buffer
    Paper: https://arxiv.org/abs/1707.01495
    Guide: https://towardsdatascience.com/hindsight-experience-replay-her-implementation-92eebab6f653
    The HER buffer uses the same trick as the standard `ReplayBuffer` using
    a single array to handle the observations. Instead of sampling and storing
    goals every time we add transitions, instead we do it all at once when
    sampling for vectorized (fast) processing.

    TODO: DOES NOT YET SUPPORT MULTIPLE ENVIRONEMTS!! More testing needed for this.

    :param env: the environment
    :param buffer_size: max number of elements in the buffer
    :param goal_selection_strategy: the goal selection strategy to be used, defaults to future
    :param n_sampled_goal: ratio of HER data to data coming from normal experience replay
    :param device: if return torch tensors on sampling, the device to attach to
    """

    def __init__(
        self,
        env: GoalEnv,
        buffer_size: int,
        goal_selection_strategy: Union[str, GoalSelectionStrategy] = "future",
        n_sampled_goal: int = 4,
        device: Union[str, T.device] = "auto",
    ) -> None:
        super().__init__(env, buffer_size, device=device)
        self.env = env
        self.desired_goals = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=env.observation_space.dtype,
        )
        self.next_achieved_goals = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=env.observation_space.dtype,
        )

        # Keep track of where in the data structure episodes end
        self.episode_end_indices = np.zeros(self.batch_shape, dtype=np.uint32)
        # Keep track of which transitions belong to which episodes.
        self.index_episode_map = np.zeros(self.batch_shape, dtype=np.uint32)
        self.episode = 0

        self._check_system_memory(
            self.observations,
            self.actions,
            self.rewards,
            self.dones,
            self.desired_goals,
            self.next_achieved_goals,
            self.index_episode_map,
        )

        if isinstance(goal_selection_strategy, str):
            self.goal_section_strategy = GoalSelectionStrategy(
                goal_selection_strategy.lower()
            )
        else:
            self.goal_section_strategy = goal_selection_strategy

        self.her_ratio = 1 - (1.0 / (n_sampled_goal + 1))

    def reset(self) -> None:
        super().reset()
        self.episode = 0

        self.desired_goals = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=self.env.observation_space.dtype,
        )
        self.next_achieved_goals = np.zeros(
            (self.buffer_size,) + self.obs_shape,
            dtype=self.env.observation_space.dtype,
        )

        # Keep track of where in the data structure episodes end
        self.episode_end_indices = np.zeros(self.batch_shape, dtype=np.int8)
        # Keep track of which transitions belong to which episodes.
        self.index_episode_map = np.zeros(self.batch_shape, dtype=np.int8)

    def add_trajectory(
        self,
        observation: Dict[str, np.ndarray],
        action: Union[np.ndarray, int],
        reward: Union[float, np.ndarray],
        next_observation: Dict[str, np.ndarray],
        done: Union[bool, np.ndarray],
    ) -> None:
        self.observations[self.pos] = observation["observation"]
        self.desired_goals[self.pos] = observation["desired_goal"]
        self.rewards[self.pos] = np.array(reward).reshape(*self.rewards.shape[1:])
        self.actions[self.pos] = np.array(action).reshape(*self.actions.shape[1:])
        self.observations[(self.pos + 1) % self.buffer_size] = next_observation[
            "observation"
        ]
        self.next_achieved_goals[self.pos] = next_observation["achieved_goal"]
        self.dones[self.pos] = np.array(done).reshape(*self.dones.shape[1:])
        self.index_episode_map[self.pos] = self.episode
        self.pos += 1

        if done:
            self.episode_end_indices[self.episode] = self.pos
            self.episode += 1

            if self.episode == self.buffer_size:
                self.episode = 0

        if self.pos == self.buffer_size:
            self.full = True
            self.pos = 0

    def _sample_goals(self, her_inds: np.ndarray) -> np.ndarray:
        """
        Sample new episode goals to calculate rewards from

        :param her_inds: the batch indices designated for new goal sampling
        :return: the new episode goals
        """
        her_episodes = self.index_episode_map[her_inds]
        # Reshaping to handle multiple environments (if self.n_envs > 1)
        episode_end_indices = self.episode_end_indices[her_episodes].reshape(
            her_episodes.shape
        )

        # Goal is the last state in the episode
        if self.goal_section_strategy == GoalSelectionStrategy.FINAL:
            goal_indices = episode_end_indices - 1

        # Goal is a random state in the same episode observed after current transition
        elif self.goal_section_strategy == GoalSelectionStrategy.FUTURE:
            # FAILURE MODE: if episode overlaps from end to beginning of buffer
            # then the randint method will likely fail due to low > high.
            # This will only happen at the overlapping episodes so quick
            # fix is to simply revert to final goal strategy in this case.
            # TODO: This will also break with multiple environments, need some sort
            # of `np.tile()` to account for this.
            if any(episode_end_indices < her_inds):
                goal_indices = episode_end_indices - 1
            else:
                goal_indices = np.random.randint(her_inds, episode_end_indices)

        else:
            raise ValueError(
                f"Strategy {self.goal_section_strategy} for samping goals not supported."
            )

        # Reshaping to handle multiple environments (if self.n_envs > 1)
        return self.next_achieved_goals[goal_indices].reshape(
            goal_indices.shape + self.obs_shape
        )

    def _sample_trajectories(
        self,
        batch_inds: np.ndarray,
    ) -> Tuple[Dict, Tensor, Tensor, Dict, Tensor]:
        """
        Get the trajectories based on batch indices calculated

        :param batch_size: number of elements to sample, included to reduce processing instead of using `len(batch_inds)`
        :param batch_inds: the indices of the elements to sample
        :param dtype: :param dtype: whether to return the trajectories as "numpy" or "torch", default numpy
        """
        her_batch_size = int(len(batch_inds) * self.her_ratio)

        # Separate HER and replay batch indices
        her_inds = batch_inds[:her_batch_size]
        replay_inds = batch_inds[her_batch_size:]

        her_goals = self._sample_goals(her_inds)
        # the new state depends on the previous state and action
        # s_{t+1} = f(s_t, a_t)
        # so the next_achieved_goal depends also on the previous state and action
        # because we are in a GoalEnv:
        # r_t = reward(s_t, a_t) = reward(next_achieved_goal, desired_goal)
        # therefore we have to use "next_achieved_goal" and not "achieved_goal"
        her_rewards = self.env.compute_reward(
            self.next_achieved_goals[her_inds], her_goals, {}
        )[:, np.newaxis]

        desired_goals = np.concatenate([her_goals, self.desired_goals[replay_inds]])
        rewards = np.concatenate([her_rewards, self.rewards[replay_inds]])
        observations = {
            "observation": self.observations[batch_inds],
            "desired_goal": desired_goals,
        }
        next_observations = {
            "observation": self.observations[(batch_inds + 1) % self.buffer_size],
            "desired_goal": desired_goals,
        }
        actions = self.actions[batch_inds]
        dones = self.dones[batch_inds]

        return observations, actions, rewards, next_observations, dones

    def sample(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> DictTrajectories:
        if isinstance(dtype, str):
            dtype = TrajectoryType(dtype.lower())

        # Sample transitions from the last complete episode recorded
        end_idx = self.episode_end_indices[self.episode - 1]
        if self.full:
            batch_inds = (
                np.random.randint(1, self.buffer_size, size=batch_size) + end_idx
            ) % self.buffer_size
        else:
            batch_inds = np.random.randint(0, end_idx, size=batch_size)

        trajectories = self._sample_trajectories(batch_inds)
        return DictTrajectories(
            observations=trajectories[0],
            actions=trajectories[1],
            rewards=trajectories[2],
            next_observations=trajectories[3],
            dones=trajectories[4],
        )

    def last(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> DictTrajectories:
        if isinstance(dtype, str):
            dtype = TrajectoryType(dtype.lower())
        assert batch_size < self.buffer_size

        # Sample transitions from the last complete episode recorded
        end_idx = self.episode_end_indices[self.episode - 1]
        start_idx = end_idx - batch_size
        if start_idx < 0 and self.full:
            batch_inds = np.concatenate((np.arange(start_idx, 0), np.arange(end_idx)))
        elif start_idx >= 0:
            batch_inds = np.arange(start_idx, end_idx)
        else:
            raise RuntimeError(
                f"Not enough samples collected, max batch_size={end_idx}"
            )

        trajectories = self._sample_trajectories(batch_inds)
        return DictTrajectories(
            observations=trajectories[0],
            actions=trajectories[1],
            rewards=trajectories[2],
            next_observations=trajectories[3],
            dones=trajectories[4],
        )

    def all(
        self, flatten_env: bool = False, dtype: Union[str, TrajectoryType] = "torch"
    ) -> DictTrajectories:
        return DictTrajectories(
            observations={
                "observation": self.observations[: self.pos],
                "desired_goal": self.desired_goals[: self.pos],
            },
            actions=self.actions[: self.pos],
            rewards=self.rewards[: self.pos],
            next_observations={
                "observation": self.observations[: (self.pos + 1) % self.buffer_size],
                "desired_goal": self.desired_goals[: (self.pos + 1) % self.buffer_size],
            },
            dones=self.dones[: self.pos],
        )



from pearll.buffers.base_buffer import BaseBuffer
from pearll.buffers.her_buffer import HERBuffer
from pearll.buffers.replay_buffer import ReplayBuffer
from pearll.buffers.rollout_buffer import RolloutBuffer

__all__ = ["BaseBuffer", "ReplayBuffer", "RolloutBuffer", "HERBuffer"]



"""Methods for estimating the Value and Q functions"""

import torch as T

from pearll.common.type_aliases import Tensor


def TD_lambda(
    rewards: Tensor,
    last_values: Tensor,
    last_dones=Tensor,
    gamma: float = 0.99,
) -> Tensor:
    """
    TD(lambda) target: https://lilianweng.github.io/lil-log/2018/02/19/a-long-peek-into-reinforcement-learning.html#temporal-difference-learning

    :param rewards: trajectory rewards
    :param last_values: last values of trajectories from a critic function (e.g. Q function, value function)
    :param last_dones: the done values of the last step of the trajectory, indicates whether to bootstrap
    :param gamma: the discount factor of future rewards
    """
    batch_size = rewards.shape[0]
    td_lambda = rewards.shape[1]
    last_dones = 1 - last_dones

    returns = T.zeros(size=(batch_size,), dtype=T.float32)
    for i in range(td_lambda):
        returns += (gamma ** i) * rewards[:, i]

    returns += (gamma ** td_lambda) * last_values * last_dones

    return returns


def TD_zero(
    rewards: Tensor,
    next_values: Tensor,
    dones: Tensor,
    gamma: float = 0.99,
) -> Tensor:
    """
    TD(0) target: https://lilianweng.github.io/lil-log/2018/02/19/a-long-peek-into-reinforcement-learning.html#combining-td-and-mc-learning

    :param rewards: trajectory rewards
    :param next_values: next values of trajectories from a critic function (e.g. Q function, value function)
    :param dones: the done values of each step of the trajectory, indicates whether to bootstrap
    :param gamma: the discount factor of future rewards
    """
    returns = rewards + ((1 - dones) * gamma * next_values)
    assert (
        returns.shape == next_values.shape
    ), f"The TD returns' shape should be {next_values.shape}, instead it is {returns.shape}."

    return returns


def soft_q_target(
    rewards: Tensor,
    dones: Tensor,
    q_values: Tensor,
    log_probs: Tensor,
    alpha: float,
    gamma: float = 0.99,
) -> Tensor:
    """
    Calculate the soft Q target: https://spinningup.openai.com/en/latest/algorithms/sac.html#id1

    :param rewards: agent reward of taking actions in the environment
    :param dones: the done values of each step of the trajectory, indicates whether to bootstrap
    :param q_values: the q value target network output
    :param log_probs: the log probability of observing the next action given the next observation
    :param alpha: entropy weighting coefficient
    :param gamma: trajectory discount
    """
    returns = rewards + gamma * (1 - dones) * (q_values - (alpha * log_probs))
    assert (
        returns.shape == q_values.shape
    ), f"The advantage's shape should be {q_values.shape}, instead it is {returns.shape}."
    return returns



from typing import Union

import numpy as np
from gym import spaces

from pearll.common.type_aliases import Observation
from pearll.common.utils import get_space_range, get_space_shape, to_numpy
from pearll.models.actor_critics import Actor, ActorCritic
from pearll.models.utils import get_mlp_size


class BaseExplorer(object):
    """
    Base explorer class that should be used when no extra noise is added to the actions/a stochastic actor is used

    :param action_space: action space
    :param start_steps: the fist n steps to uniformally sample actions
    """

    def __init__(
        self,
        action_space: spaces.Space,
        start_steps: int = 20000,
    ) -> None:
        self.start_steps = start_steps
        self.action_space = action_space
        self.action_shape = get_space_shape(action_space)
        self.action_size = get_mlp_size(get_space_shape(action_space))
        self.action_range = get_space_range(action_space)

    def __call__(
        self, model: Union[Actor, ActorCritic], observation: Observation, step: int
    ) -> np.ndarray:
        """
        Get an action for the given observation in training.

        :param model: the model to use
        :param observation: the observation
        :param step: the current training step
        """
        if step >= self.start_steps:
            action = to_numpy(model(observation))
            action = np.clip(action, self.action_range[0], self.action_range[1])
        else:
            action = self.action_space.sample()

        return action



from abc import ABC, abstractmethod
from typing import Iterator, Type, Union

import torch as T
from torch.nn.parameter import Parameter

from pearll.common.type_aliases import UpdaterLog
from pearll.models.actor_critics import ActorCritic, Critic


class BaseCriticUpdater(ABC):
    """
    The base class with pre-defined methods for derived classes

    :param loss_class: The loss class to use e.g. MSE
    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        loss_class: T.nn.Module = T.nn.MSELoss(),
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        self.loss_class = loss_class
        self.optimizer_class = optimizer_class
        self.max_grad = max_grad

    def _get_model_parameters(
        self, model: Union[Critic, ActorCritic]
    ) -> Iterator[Parameter]:
        """Get the critic model parameters"""
        if isinstance(model, Critic):
            return model.parameters()
        else:
            params = []
            for critic in model.critics:
                params.extend(critic.model.parameters())
            return params

    def run_optimizer(
        self,
        optimizer: T.optim.Optimizer,
        loss: T.Tensor,
        critic_parameters: Iterator[Parameter],
    ) -> None:
        """Run an optimization step"""
        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        if self.max_grad > 0:
            T.nn.utils.clip_grad_norm_(critic_parameters, self.max_grad)
        optimizer.step()

    @abstractmethod
    def __call__(self, model: Union[Critic, ActorCritic]) -> UpdaterLog:
        """Run an optimization step"""


class ValueRegression(BaseCriticUpdater):
    """
    Regression for a value function estimator

    :param loss_class: the distance loss class for regression, defaults to MSE
    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        loss_class: T.nn.Module = T.nn.MSELoss(),
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(
            loss_class=loss_class, optimizer_class=optimizer_class, max_grad=max_grad
        )

    def __call__(
        self,
        model: Union[Critic, ActorCritic],
        observations: T.Tensor,
        returns: T.Tensor,
        learning_rate: float = 0.001,
        loss_coeff: float = 1,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: the model on which the optimization should be run
        :param observations: observation inputs
        :param returns: the target to regress to (e.g. TD Values, Monte-Carlo Values)
        :param learning_rate: the learning rate for the optimizer algorithm
        :param loss_coeff: the coefficient for the value loss, defaults to 1
        """
        critic_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(critic_parameters, lr=learning_rate)

        if isinstance(model, Critic):
            values = model(observations)
        else:
            values = model.forward_critics(observations)

        loss = loss_coeff * self.loss_class(values, returns)

        self.run_optimizer(optimizer, loss, critic_parameters)

        return UpdaterLog(loss=loss.detach())


class ContinuousQRegression(BaseCriticUpdater):
    """
    Regression for a continuous Q function estimator

    :param loss_class: the distance loss class for regression, defaults to MSE
    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        loss_class: T.nn.Module = T.nn.MSELoss(),
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(
            loss_class=loss_class, optimizer_class=optimizer_class, max_grad=max_grad
        )

    def __call__(
        self,
        model: Union[Critic, ActorCritic],
        observations: T.Tensor,
        actions: T.Tensor,
        returns: T.Tensor,
        learning_rate: float = 0.001,
        loss_coeff: float = 1,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: the model on which the optimization should be run
        :param observations: observation inputs
        :param actions: action inputs needed for continuous Q function modelling
        :param returns: the target to regress to (e.g. TD Values, Monte-Carlo Values)
        :param learning_rate: the learning rate for the optimizer algorithm
        :param loss_coeff: the coefficient for the Q loss, defaults to 1
        """
        critic_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(critic_parameters, lr=learning_rate)

        if isinstance(model, Critic):
            q_values = model(observations, actions)
        else:
            q_values = model.forward_critics(observations, actions)

        loss = loss_coeff * self.loss_class(q_values, returns)

        self.run_optimizer(optimizer, loss, critic_parameters)

        return UpdaterLog(loss=loss.detach())


class DiscreteQRegression(BaseCriticUpdater):
    """
    Regression for a discrete Q function estimator

    :param loss_class: the distance loss class for regression, defaults to MSE
    :param optimizer_class: the type of optimizer to use, defaults to Adam
    :param max_grad: maximum gradient clip value, defaults to no clipping with a value of 0
    """

    def __init__(
        self,
        loss_class: T.nn.Module = T.nn.MSELoss(),
        optimizer_class: Type[T.optim.Optimizer] = T.optim.Adam,
        max_grad: float = 0,
    ) -> None:
        super().__init__(
            loss_class=loss_class, optimizer_class=optimizer_class, max_grad=max_grad
        )

    def __call__(
        self,
        model: Union[Critic, ActorCritic],
        observations: T.Tensor,
        returns: T.Tensor,
        actions_index: T.Tensor,
        learning_rate: float = 0.001,
        loss_coeff: float = 1,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param model: the model on which the optimization should be run
        :param observations: observation inputs
        :param returns: the target to regress to (e.g. TD Values, Monte-Carlo Values)
        :param actions_index: discrete action values to use as indices, needed to filter
            the Q values for actions experienced.
        :param learning_rate: the learning rate for the optimizer algorithm
        :param loss_coeff: the coefficient for the Q loss, defaults to 1
        """
        critic_parameters = self._get_model_parameters(model)
        optimizer = self.optimizer_class(critic_parameters, lr=learning_rate)

        if isinstance(model, Critic):
            q_values = model(observations)
        else:
            q_values = model.forward_critics(observations)
        q_values = T.gather(q_values, dim=-1, index=actions_index.long())

        loss = loss_coeff * self.loss_class(q_values, returns)

        self.run_optimizer(optimizer, loss, critic_parameters)

        return UpdaterLog(loss=loss.detach().item())



from typing import Union

import numpy as np
import torch as T
from gym import Env

from pearll.buffers.base_buffer import BaseBuffer
from pearll.common.enumerations import TrajectoryType
from pearll.common.type_aliases import Trajectories


class ReplayBuffer(BaseBuffer):
    """
    Replay buffer handles sample collection and processing for off-policy algorithms.
    We use a single array to save space in handling observations and next observations
    rather than using two different arrays. This assumes observations are stored sequentially
    and that observations can be 'blended' only if done = True since the next_observation
    in this case is essentially discarded anyway in most objective functions (at the end of
    the trajectory the value or q function of the terminal state is always 0!).

    :param env: the environment
    :param buffer_size: max number of elements in the buffer
    """

    def __init__(
        self,
        env: Env,
        buffer_size: int,
        device: Union[str, T.device] = "auto",
    ) -> None:
        super().__init__(
            env,
            buffer_size,
            device,
        )
        self._check_system_memory(
            self.observations, self.actions, self.rewards, self.dones
        )

    def reset(self) -> None:
        super().reset()

    def add_trajectory(
        self,
        observation: np.ndarray,
        action: Union[np.ndarray, int],
        reward: Union[float, np.ndarray],
        next_observation: np.ndarray,
        done: Union[bool, np.ndarray],
    ) -> None:
        self.observations[self.pos] = observation
        self.observations[(self.pos + 1) % self.buffer_size] = next_observation
        self.actions[self.pos] = np.array(action).reshape(*self.actions.shape[1:])
        self.rewards[self.pos] = np.array(reward).reshape(*self.rewards.shape[1:])
        self.dones[self.pos] = np.array(done).reshape(*self.dones.shape[1:])

        self.pos += 1
        if self.pos == self.buffer_size:
            self.full = True
            self.pos = 0

    def sample(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        if self.full:
            batch_inds = (
                np.random.randint(1, self.buffer_size, size=batch_size) + self.pos
            ) % self.buffer_size
        else:
            batch_inds = np.random.randint(0, self.pos, size=batch_size)

        observations = self.observations[batch_inds]
        actions = self.actions[batch_inds]
        rewards = self.rewards[batch_inds]
        next_observations = self.observations[(batch_inds + 1) % self.buffer_size]
        dones = self.dones[batch_inds]

        return self._transform_samples(
            flatten_env, dtype, observations, actions, rewards, next_observations, dones
        )

    def last(
        self,
        batch_size: int,
        flatten_env: bool = False,
        dtype: Union[str, TrajectoryType] = "torch",
    ) -> Trajectories:
        assert batch_size < self.buffer_size

        start_idx = self.pos - batch_size
        if start_idx < 0:
            batch_inds = np.concatenate((np.arange(start_idx, 0), np.arange(self.pos)))
        else:
            batch_inds = np.arange(start_idx, self.pos)

        observations = self.observations[batch_inds]
        actions = self.actions[batch_inds]
        rewards = self.rewards[batch_inds]
        next_observations = self.observations[(batch_inds + 1) % self.buffer_size]
        dones = self.dones[batch_inds]

        return self._transform_samples(
            flatten_env, dtype, observations, actions, rewards, next_observations, dones
        )

    def all(
        self, flatten_env: bool = False, dtype: Union[str, TrajectoryType] = "torch"
    ) -> Trajectories:
        return self._transform_samples(
            flatten_env=flatten_env,
            dtype=dtype,
            observations=self.observations[: self.pos],
            actions=self.actions[: self.pos],
            rewards=self.rewards[: self.pos],
            next_observations=self.observations[: (self.pos + 1) % self.buffer_size],
            dones=self.dones[: self.pos],
        )



"""Methods for generating a new population from selected individuals"""

from typing import Optional, Tuple

import numpy as np


def fit_gaussian(parents: np.ndarray, population_shape: Tuple[int, ...]) -> np.ndarray:
    """
    Generates a new population given selected parents fitted to a Gaussian distribution.

    :param parents: the parent population
    :param population_shape: the shape of the new population
    :return: the new population
    """
    # Get the mean and standard deviation of the parents
    mean = np.mean(parents, axis=0)
    std = np.std(parents, axis=0)

    return np.random.normal(mean, std, size=population_shape)


def one_point_crossover(
    parents: np.ndarray,
    crossover_index: Optional[int] = None,
) -> np.ndarray:
    """
    Generates a new population from two parent individuals using one-point crossover.
    https://www.tutorialspoint.com/genetic_algorithms/genetic_algorithms_crossover.htm

    :param parents: the parent population
    :param crossover_index: crossover point index, if None then randomly selected
    :return: the new population
    """
    # Split population into parent pairs
    pairs = np.array([[a, b] for a, b in zip(parents[::2], parents[1::2])])

    # Get crossover indices
    if crossover_index is None:
        crossover_indices = np.random.choice(
            np.arange(parents.shape[1]),
            size=pairs.shape[0],
        )
    else:
        crossover_indices = np.full(pairs.shape[0], crossover_index)

    # Perform crossover
    for i, pair in enumerate(pairs):
        pairs[i] = [
            np.concatenate(
                (pair[0][: crossover_indices[i]], pair[1][crossover_indices[i] :]),
                axis=0,
            ),
            np.concatenate(
                (pair[1][: crossover_indices[i]], pair[0][crossover_indices[i] :]),
                axis=0,
            ),
        ]
    new_population = np.concatenate(pairs, axis=0)

    # Account for odd parent population size
    if parents.shape[0] % 2 != 0:
        new_population = np.concatenate((new_population, [parents[-1]]), axis=0)

    return new_population



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pearll',
 'pearll.agents',
 'pearll.buffers',
 'pearll.callbacks',
 'pearll.common',
 'pearll.explorers',
 'pearll.models',
 'pearll.signal_processing',
 'pearll.updaters']

package_data = \
{'': ['*']}

install_requires = \
['gym>=0.19.0,<0.20.0',
 'matplotlib>=3.5.0,<4.0.0',
 'numpy>=1.20.3,<2.0.0',
 'psutil>=5.8.0,<6.0.0',
 'scikit-learn>=1.0.1,<2.0.0',
 'tensorboard>=2.7.0,<3.0.0',
 'torch>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'pearll',
    'version': '0.3.2',
    'description': 'Adaptable tools to make reinforcement learning and evolutionary computation algorithms',
    'long_description': None,
    'author': 'Rohan Tangri',
    'author_email': 'rohan.tangri@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



import logging
from typing import Optional, Union

import numpy as np
from torch.utils.tensorboard import SummaryWriter

from pearll.common.type_aliases import Log


def get_logger(file_handler_level: int, stream_handler_level: int) -> logging.Logger:
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    if logger.hasHandlers():
        logger.handlers.clear()
    file_handler = logging.FileHandler("agent.log", mode="w")
    file_handler.setLevel(file_handler_level)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(stream_handler_level)

    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)

    return logger


class Logger(object):
    """
    The Logger object combines the torch SummaryWriter with python in-built logging

    :param tensorboard_log_path: path to store the tensorboard log
    :param file_handler_level: logging level for the file log
    :param stream_handeler_level: logging level for the streaming log
    :param verbose: whether to display at all or not
    :param num_envs: number of environments to run, useful for multi-agent
    """

    def __init__(
        self,
        tensorboard_log_path: Optional[str] = None,
        file_handler_level: int = logging.DEBUG,
        stream_handler_level: int = logging.INFO,
        verbose: bool = True,
        num_envs: int = 1,
    ) -> None:
        self.writer = SummaryWriter(tensorboard_log_path)
        self.logger = get_logger(file_handler_level, stream_handler_level)
        self.verbose = verbose
        self.num_envs = num_envs
        self.actor_losses = []
        self.critic_losses = []
        self.divergences = []
        self.entropies = []
        self.rewards = []
        # Keep track of which environments have completed an episode
        self.episode_dones = np.array([False for _ in range(num_envs)])

    def reset_log(self) -> None:
        self.actor_losses = []
        self.critic_losses = []
        self.divergences = []
        self.entropies = []
        self.rewards = []
        self.episode_dones = np.array([False for _ in range(self.num_envs)])

    def add_train_log(self, train_log: Log) -> None:
        if train_log.actor_loss is not None:
            self.actor_losses.append(train_log.actor_loss)
        if train_log.critic_loss is not None:
            self.critic_losses.append(train_log.critic_loss)
        if train_log.entropy is not None:
            self.entropies.append(train_log.entropy)
        if train_log.divergence is not None:
            self.divergences.append(train_log.divergence)

    def add_reward(self, reward: Union[float, np.ndarray, int]) -> None:
        """Add step reward to the episode rewards"""
        if isinstance(reward, (float, np.floating, int)):
            self.rewards.append(reward)
        elif isinstance(reward, np.ndarray):
            log_reward = reward[np.where(self.episode_dones == False)[0]].mean()
            if not np.isnan(log_reward):
                self.rewards.append(log_reward)
        else:
            raise TypeError(
                f"Reward must be a number or numpy array, got {type(reward)}"
            )

    def check_episode_done(self, done: np.ndarray) -> bool:
        """
        Check if all the environments have completed an episode

        :param done: done array from the environment
        """
        self.episode_dones = np.logical_or(self.episode_dones, done)
        return np.all(self.episode_dones)

    def _make_episode_log(self) -> Log:
        """Make an episode log out of the collected stats"""
        episode_log = Log(
            reward=np.sum(self.rewards),
        )
        if self.actor_losses:
            episode_log.actor_loss = np.mean(self.actor_losses)
        if self.critic_losses:
            episode_log.critic_loss = np.mean(self.critic_losses)
        if self.divergences:
            episode_log.divergence = np.mean(self.divergences)
        if self.entropies:
            episode_log.entropy = np.mean(self.entropies)

        return episode_log

    def write_log(self, step: int) -> None:
        """Write a log to tensorboard and python logging"""
        episode_log = self._make_episode_log()
        self.writer.add_scalar("Reward/episode_reward", episode_log.reward, step)
        if episode_log.actor_loss is not None:
            self.writer.add_scalar("Loss/actor_loss", episode_log.actor_loss, step)
        if episode_log.critic_loss is not None:
            self.writer.add_scalar("Loss/critic_loss", episode_log.critic_loss, step)
        if episode_log.divergence is not None:
            self.writer.add_scalar("Metrics/divergence", episode_log.divergence, step)
        if episode_log.entropy is not None:
            self.writer.add_scalar("Metrics/entropy", episode_log.entropy, step)

        if self.verbose:
            self.logger.info(f"{step}: {episode_log}")

    def info(self, msg: str):
        if self.verbose:
            self.logger.info(msg)

    def debug(self, msg: str):
        if self.verbose:
            self.logger.debug(msg)

    def warning(self, msg: str):
        if self.verbose:
            self.logger.warning(msg)

    def error(self, msg: str):
        if self.verbose:
            self.logger.error(msg)

    def exception(self, msg: str):
        if self.verbose:
            self.logger.exception(msg)



import os
from enum import Enum
from typing import Any, List, Optional, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from tensorboard.backend.event_processing import event_accumulator


class Metric(Enum):
    """
    Enum for the different plots.
    """

    REWARD = "reward"
    ACTOR_LOSS = "actor"
    CRITIC_LOSS = "critic"
    DIVERGENCE = "divergence"
    ENTROPY = "entropy"


def read_tensorboard_data(
    path: str, metric: str
) -> List[event_accumulator.ScalarEvent]:
    """
    Reads tensorboard data from a given path.
    :param path: path to the tensorboard data
    :param plot: the plot to read
    """
    metric = Metric(metric.lower())
    if metric == Metric.REWARD:
        tag = "Reward/episode_reward"
    elif metric == Metric.ACTOR_LOSS:
        tag = "Loss/actor_loss"
    elif metric == Metric.CRITIC_LOSS:
        tag = "Loss/critic_loss"
    elif metric == Metric.DIVERGENCE:
        tag = "Metrics/divergence"
    elif metric == Metric.ENTROPY:
        tag = "Metrics/entropy"

    ea = event_accumulator.EventAccumulator(
        path,
        size_guidance={
            event_accumulator.COMPRESSED_HISTOGRAMS: 500,
            event_accumulator.IMAGES: 4,
            event_accumulator.AUDIO: 4,
            event_accumulator.SCALARS: 0,
            event_accumulator.HISTOGRAMS: 1,
        },
    )

    ea.Reload()
    return ea.Scalars(tag)


def get_axis_data(
    data: List[event_accumulator.ScalarEvent],
    x_axis: str = "step",
    y_axis: str = "value",
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Gets the axis data from the given data.
    :param data: the data to get the axis data from
    :param x_axis: the x axis tag
    :param y_axis: the y axis tag
    """
    if x_axis == "step":
        x_data = [event.step for event in data]
    elif x_axis == "wall_time":
        x_data = [event.wall_time for event in data]
    elif x_axis == "value":
        x_data = [event.value for event in data]

    if y_axis == "value":
        y_data = [event.value for event in data]
    elif y_axis == "wall_time":
        y_data = [event.wall_time for event in data]
    elif y_axis == "step":
        y_data = [event.step for event in data]

    return np.array(x_data), np.array(y_data)


def get_file_name(path: str) -> str:
    """
    Gets the file name from the given path.
    :param path: the path to the file
    """
    return os.path.basename(path).split(".")[0]


def smooth(vals: np.ndarray, window: int) -> np.ndarray:
    """Smooths values using a sliding window."""

    if window > 1:
        if window > len(vals):
            window = len(vals)
        y = np.ones(window)
        x = vals
        z = np.ones(len(vals))
        mode = "same"
        vals = np.convolve(x, y, mode) / np.convolve(z, y, mode)

    return vals


def stats(
    time_series: np.ndarray, window: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Extract rolling statistics over the common range of x values.

    :param time_series: the time series to extract the rolling statistics from
    :param window: the window size
    """
    windowed_series = np.lib.stride_tricks.sliding_window_view(time_series, window)
    std = np.std(windowed_series, axis=1)
    min = np.min(windowed_series, axis=1)
    max = np.max(windowed_series, axis=1)
    std = np.concatenate((std, np.std(windowed_series[-window + 1 :], axis=1)))
    min = np.concatenate((min, np.min(windowed_series[-window + 1 :], axis=1)))
    max = np.concatenate((max, np.max(windowed_series[-window + 1 :], axis=1)))
    return std, min, max


def plot(
    paths: List[List[str]],
    metric: str,
    titles: List[str],
    num_cols: int,
    interval: str = "bounds",
    legend: Optional[List[str]] = None,
    window: int = 0,
    xlabel: str = "Step",
    ylabel: str = "Reward",
    x_axis: str = "step",
    y_axis: str = "value",
    log_y: bool = False,
    save_types: List[str] = ["pdf"],
    save_path: Optional[str] = os.path.join(os.getcwd(), "plots/plot"),
) -> None:
    """
    Plots the given data.
    :param paths: the paths to the tensorboard data
    :param metric: the metric to plot
    :param titles: the titles of the plots
    :param num_cols: the number of columns to use
    :param interval: the interval to use
    :param legend: the legend to use
    :param window: the window size
    :param xlabel: the x axis tag
    :param ylabel: the y axis tag
    :param x_axis: the x axis data
    :param y_axis: the y axis data
    :param log_y: whether to log the y axis
    :param save_types: the save types
    :param save_path: where to save the plots
    """
    num_plots = len(paths)
    num_rows = int(np.ceil(num_plots / num_cols))
    plt.ion()
    fig = plt.figure(figsize=(num_cols * 6, num_rows * 5))
    fig.clear()
    grid = mpl.gridspec.GridSpec(
        nrows=num_rows + 1,
        ncols=1 + num_cols,
        hspace=0.3,
        height_ratios=[1] * num_rows + [0.1],
        width_ratios=[0] + [1] * num_cols,
    )
    axes = []
    for i in range(num_plots):
        ax = fig.add_subplot(grid[i // num_cols, 1 + i % num_cols])
        axes.append(ax)

    for collection, ax, title in zip(paths, axes, titles):
        agents = [get_file_name(path) for path in collection]
        num_agents = len(agents)

        if num_agents <= 10:
            cmap = "tab10"
        elif num_agents <= 20:
            cmap = "tab20"
        else:
            cmap = "rainbow"
        cmap = plt.get_cmap(cmap)

        if isinstance(cmap, mpl.colors.ListedColormap):
            colors = cmap(range(num_agents))
        else:
            colors = list(cmap(np.linspace(0, 1, num_agents)))
        agent_colors = {a: c for a, c in zip(agents, colors)}

        for path, agent in zip(collection, agents):
            data = read_tensorboard_data(path, metric)
            x_data, y_data = get_axis_data(data, x_axis, y_axis)
            mean = smooth(y_data, window=window)
            label = agent if legend is None else legend[agents.index(agent)]
            ax.plot(x_data, mean, label=label, color=agent_colors[agent], alpha=1, lw=2)
            if window > 0:
                std, min_, max_ = stats(y_data, window)
                if interval in ["std", "bounds"]:
                    if interval == "std":
                        ax.fill_between(
                            x_data,
                            mean - std,
                            mean + std,
                            alpha=0.1,
                            color=agent_colors[agent],
                            lw=0,
                        )
                    elif interval == "bounds":
                        ax.fill_between(
                            x_data,
                            min_,
                            max_,
                            alpha=0.1,
                            color=agent_colors[agent],
                            lw=0,
                        )

        # Finalize the figures.
        ax.locator_params(axis="x", nbins=6)
        ax.locator_params(axis="y", tight=True, nbins=6)
        ax.get_yaxis().set_major_formatter(
            mpl.ticker.FuncFormatter(lambda x, p: f"{x:,g}")
        )
        low, high = ax.get_xlim()
        if max(abs(low), abs(high)) >= 1e3:
            ax.ticklabel_format(style="sci", axis="x", scilimits=(0, 0))
        ax.xaxis.grid(linewidth=0.5, alpha=0.5)
        ax.yaxis.grid(linewidth=0.5)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.tick_params(axis="both", which="both", length=0)
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend()
        if log_y:
            ax.set_yscale("log")

    if save_path:
        if not os.path.exists(os.path.dirname(save_path)):
            os.makedirs(os.path.dirname(save_path))
        for save_type in save_types:
            plt.savefig(f"{save_path}.{save_type}", bbox_inches="tight")

    plt.show()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument(
        "-p", "--paths", nargs="+", type=str, action="append", required=True
    )
    parser.add_argument("--metric", type=str, required=True)
    parser.add_argument("--titles", nargs="+", type=str, required=True)
    parser.add_argument("--num-cols", type=int, default=1)
    parser.add_argument("--interval", type=str, default="bounds")
    parser.add_argument("--legend", nargs="+", type=str)
    parser.add_argument("--window", type=int, default=0)
    parser.add_argument("--xlabel", type=str, default="Step")
    parser.add_argument("--ylabel", type=str, default="Reward")
    parser.add_argument("--x-axis", type=str, default="step")
    parser.add_argument("--y-axis", type=str, default="value")
    parser.add_argument("--log-y", action="store_true")
    parser.add_argument("--save-types", nargs="+", type=str, default=["pdf"])
    parser.add_argument(
        "--save-path", type=str, default=os.path.join(os.getcwd(), "plots/plot")
    )

    args = parser.parse_args()

    plot(
        args.paths,
        args.metric,
        args.titles,
        args.num_cols,
        args.interval,
        args.legend,
        args.window,
        args.xlabel,
        args.ylabel,
        args.x_axis,
        args.y_axis,
        args.log_y,
        args.save_types,
        args.save_path,
    )



from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
from gym.spaces import Discrete, MultiDiscrete
from torch.distributions import Normal, kl_divergence

from pearll.common.type_aliases import (
    CrossoverFunc,
    MutationFunc,
    SelectionFunc,
    UpdaterLog,
)
from pearll.common.utils import to_torch
from pearll.models.actor_critics import ActorCritic


class BaseEvolutionUpdater(ABC):
    """
    The base random search updater class with pre-defined methods for derived classes

    :param model: the actor critic model containing the population
    :param population_type: the type of population to update, either "actor" or "critic"
    """

    def __init__(self, model: ActorCritic, population_type: str = "actor") -> None:
        self.model = model
        self.population_type = population_type
        if population_type == "actor":
            self.mean = model.mean_actor
            self.std = self.model.population_settings.actor_std
            self.population_size = model.num_actors
            self.normal_dist = model.normal_dist_actor
            self.space_shape = model.actor.space_shape
            self.space_range = model.actor.space_range
            self.space = model.actor.space
        elif population_type == "critic":
            self.mean = model.mean_critic
            self.std = self.model.population_settings.critic_std
            self.population_size = model.num_critics
            self.normal_dist = model.normal_dist_critic
            self.space_shape = model.critic.space_shape
            self.space_range = model.critic.space_range
            self.space = model.critic.space

    def update_networks(self, population: np.ndarray) -> None:
        """
        Update the networks in the population

        :param population: the population state to set the networks to
        """
        if self.population_type == "actor":
            self.model.set_actors_state(population)
        elif self.population_type == "critic":
            self.model.set_critics_state(population)

    @abstractmethod
    def __call__(self) -> UpdaterLog:
        """Run an optimization step"""


class NoisyGradientAscent(BaseEvolutionUpdater):
    """
    Updater for the Natural Evolutionary Strategy

    :param model: the actor critic model containing the population
    :param population_type: the type of population to update, either "actor" or "critic"
    """

    def __init__(self, model: ActorCritic, population_type: str = "actor") -> None:
        super().__init__(model, population_type)

    def __call__(
        self,
        learning_rate: float,
        optimization_direction: np.ndarray,
        mutation_operator: Optional[MutationFunc] = None,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param learning_rate: the learning rate
        :param optimization_direction: the optimization direction
        :param mutation_operator: the mutation operator
        :return: the updater log
        """
        # Snapshot current population dist for kl divergence
        # use copy() to avoid modifying the original
        old_dist = Normal(to_torch(self.mean.copy()), self.std)

        # Main update
        self.mean += learning_rate * optimization_direction

        # Generate new population
        self.normal_dist = np.random.randn(self.population_size, *self.space_shape)
        population = self.mean + (self.std * self.normal_dist)
        if mutation_operator is not None:
            population = mutation_operator(population, self.space)

        # Discretize and clip population as needed
        if isinstance(self.space, (Discrete, MultiDiscrete)):
            population = np.round(population).astype(np.int32)
        population = np.clip(population, self.space_range[0], self.space_range[1])
        self.update_networks(population)

        # Calculate Log metrics
        new_dist = Normal(to_torch(self.mean), self.std)
        population_entropy = new_dist.entropy().mean()
        population_kl = kl_divergence(old_dist, new_dist).mean()

        return UpdaterLog(divergence=population_kl, entropy=population_entropy)


class GeneticUpdater(BaseEvolutionUpdater):
    """
    Updater for the Genetic Algorithm

    :param model: the actor critic model containing the population
    :param population_type: the type of population to update, either "actor" or "critic"
    """

    def __init__(self, model: ActorCritic, population_type: str = "actor") -> None:
        super().__init__(model, population_type)

    def __call__(
        self,
        rewards: np.ndarray,
        selection_operator: Optional[SelectionFunc] = None,
        crossover_operator: Optional[CrossoverFunc] = None,
        mutation_operator: Optional[MutationFunc] = None,
        elitism: float = 0.1,
    ) -> UpdaterLog:
        """
        Perform an optimization step

        :param rewards: the rewards for the current population
        :param selection_operator: the selection operator function
        :param crossover_operator: the crossover operator function
        :param mutation_operator: the mutation operator function
        :param elitism: fraction of the population to keep as elite
        :return: the updater log
        """
        # Store elite population
        if self.population_type == "actor":
            old_population = self.model.numpy_actors()
        elif self.population_type == "critic":
            old_population = self.model.numpy_critics()
        if elitism > 0:
            num_elite = int(self.population_size * elitism)
            elite_indices = np.argpartition(rewards, -num_elite)[-num_elite:]
            elite_population = old_population[elite_indices]

        # Main update
        if selection_operator is not None:
            new_population = selection_operator(old_population, rewards)
        if crossover_operator is not None:
            new_population = crossover_operator(new_population)
        if mutation_operator is not None:
            new_population = mutation_operator(new_population, self.space)
        if elitism > 0:
            new_population[elite_indices] = elite_population
        self.update_networks(new_population)

        # Calculate Log metrics
        divergence = np.mean(np.abs(new_population - old_population))
        entropy = np.mean(
            np.abs(np.max(new_population, axis=0) - np.min(new_population, axis=0))
        )

        return UpdaterLog(divergence=divergence, entropy=entropy)



from functools import partial
from typing import Callable, List, Optional, Type

import numpy as np
from gym.vector.vector_env import VectorEnv

from pearll.agents.base_agents import BaseAgent
from pearll.buffers import RolloutBuffer
from pearll.buffers.base_buffer import BaseBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import filter_rewards
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models import ActorCritic, Dummy
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    MutationSettings,
    PopulationSettings,
    Settings,
)
from pearll.signal_processing import (
    crossover_operators,
    mutation_operators,
    selection_operators,
)
from pearll.updaters.evolution import BaseEvolutionUpdater, GeneticUpdater


def default_model(env: VectorEnv):
    """
    Returns a default model for the given environment.
    """
    actor = Dummy(space=env.single_action_space)
    critic = Dummy(space=env.single_action_space)

    return ActorCritic(
        actor=actor,
        critic=critic,
        population_settings=PopulationSettings(
            actor_population_size=env.num_envs, actor_distribution="uniform"
        ),
    )


class GA(BaseAgent):
    """
    Genetic Algorithm
    https://www.geeksforgeeks.org/genetic-algorithms/

    :param env: the gym-like environment to be used, should be a VectorEnv
    :param model: the neural network model
    :param updater_class: the updater class to be used
    :param selection_operator: the selection operator to be used
    :param selection_settings: the selection settings to be used
    :param crossover_operator: the crossover operator to be used
    :param crossover_settings: the crossover settings to be used
    :param mutation_operator: the mutation operator to be used
    :param mutation_settings: the mutation settings to be used
    :param elitism: the elitism ratio
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: VectorEnv,
        model: Optional[ActorCritic] = None,
        updater_class: Type[BaseEvolutionUpdater] = GeneticUpdater,
        selection_operator: Callable = selection_operators.roulette_selection,
        selection_settings: Settings = Settings(),
        crossover_operator: Callable = crossover_operators.one_point_crossover,
        crossover_settings: Settings = Settings(),
        mutation_operator: Callable = mutation_operators.uniform_mutation,
        mutation_settings: MutationSettings = MutationSettings(),
        elitism: float = 0.1,
        buffer_class: Type[BaseBuffer] = RolloutBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=0),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model if model is not None else default_model(env)
        super().__init__(
            env=env,
            model=model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )

        self.updater = updater_class(self.model)

        self.selection_operator = partial(
            selection_operator, **selection_settings.filter_none()
        )
        self.crossover_operator = partial(
            crossover_operator, **crossover_settings.filter_none()
        )
        self.mutation_operator = partial(
            mutation_operator, **mutation_settings.filter_none()
        )
        self.elitism = elitism

    def _fit(
        self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1
    ) -> Log:
        divergences = np.zeros(actor_epochs)
        entropies = np.zeros(actor_epochs)

        trajectories = self.buffer.sample(batch_size, dtype="numpy")
        rewards = trajectories.rewards.squeeze()
        rewards = filter_rewards(rewards, trajectories.dones.squeeze())
        if rewards.ndim > 1:
            rewards = rewards.sum(dim=-1)
        for i in range(actor_epochs):
            log = self.updater(
                rewards=rewards,
                selection_operator=self.selection_operator,
                crossover_operator=self.crossover_operator,
                mutation_operator=self.mutation_operator,
                elitism=self.elitism,
            )
            divergences[i] = log.divergence
            entropies[i] = log.entropy
        self.buffer.reset()

        return Log(divergence=divergences.sum(), entropy=entropies.mean())



from typing import Dict, List, Optional, Type

import numpy as np
import torch as T
from gym import spaces

from pearll.common.type_aliases import Tensor
from pearll.common.utils import to_numpy
from pearll.models.utils import concat_obs_actions


class IdentityEncoder(T.nn.Module):
    """This encoder passes the input through unchanged."""

    def __init__(self):
        super().__init__()

    def forward(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        # Some algorithms use both the observations and actions as input (e.g. DDPG for conitnuous Q function)
        observations = concat_obs_actions(observations, actions)
        return observations


class FlattenEncoder(T.nn.Module):
    """This encoder flattens the input."""

    def __init__(self):
        super().__init__()

    def forward(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        # Some algorithms use both the observations and actions as input (e.g. DDPG for conitnuous Q function)
        # Make sure observations is a torch tensor, get error if numpy for some reason??
        observations = concat_obs_actions(observations, actions)
        return T.flatten(observations)


class MLPEncoder(T.nn.Module):
    """This is a single layer MLP encoder"""

    def __init__(self, input_size, output_size):
        super().__init__()
        self.model = T.nn.Linear(input_size, output_size)

    def forward(
        self, observations: Tensor, actions: Optional[Tensor] = None
    ) -> T.Tensor:
        observations = concat_obs_actions(observations, actions)
        return self.model(observations)


class CNNEncoder(T.nn.Module):
    """
    CNN from DQN nature paper:
    Mnih, Volodymyr, et al.
    "Human-level control through deep reinforcement learning."
    Nature 518.7540 (2015): 529-533.

    :param observation_space:
    :param output_size: number neurons in the last layer.
    :param activation_fn: the activation function after each layer
    """

    def __init__(
        self,
        observation_space: spaces.Box,
        output_size: int = 512,
        activation_fn: Type[T.nn.Module] = T.nn.ReLU,
    ):
        super().__init__()
        # We assume CxHxW images (channels first)
        # Re-ordering will be done by pre-preprocessing or wrapper
        n_input_channels = observation_space.shape[0]
        self.cnn = T.nn.Sequential(
            T.nn.Conv2d(n_input_channels, 32, kernel_size=8, stride=4, padding=0),
            activation_fn(),
            T.nn.Conv2d(32, 64, kernel_size=4, stride=2, padding=0),
            activation_fn(),
            T.nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=0),
            activation_fn(),
            T.nn.Flatten(),
        )

        # Compute shape by doing one forward pass
        with T.no_grad():
            n_flatten = self.cnn(
                T.as_tensor(observation_space.sample()[None]).float()
            ).shape[1]

        self.linear = T.nn.Sequential(T.nn.Linear(n_flatten, output_size), T.nn.ReLU())

    def forward(self, observations: Tensor) -> T.Tensor:
        return self.linear(self.cnn(observations))


class DictEncoder(T.nn.Module):
    """
    Handles dictionary observations, e.g. from GoalEnv

    :param labels: dictionary labels to extract for model
    :param encoder: encoder module to run after extracting array from dictionary
    """

    def __init__(
        self,
        labels: List[str] = ["observation", "desired_goal"],
        encoder: T.nn.Module = IdentityEncoder(),
    ) -> None:
        super().__init__()
        self.labels = labels
        self.encoder = encoder

    def forward(
        self, observations: Dict[str, Tensor], actions: Optional[Tensor] = None
    ) -> T.Tensor:
        obs = [observations[label] for label in self.labels]
        obs = to_numpy(*obs)
        if len(self.labels) > 1:
            shape_length = len(observations[self.labels[0]].shape)
            obs = np.concatenate(obs, axis=shape_length - 1)
        return self.encoder(obs, actions)



from typing import List, Optional, Type

import gym
import numpy as np
import torch as T
from gym import Env

from pearll.agents.base_agents import BaseAgent
from pearll.buffers import BaseBuffer, RolloutBuffer
from pearll.callbacks.base_callback import BaseCallback
from pearll.common.type_aliases import Log
from pearll.common.utils import get_space_shape
from pearll.explorers import BaseExplorer
from pearll.models import Actor, ActorCritic, Critic
from pearll.models.encoders import IdentityEncoder
from pearll.models.heads import CategoricalHead, ValueHead
from pearll.models.torsos import MLP
from pearll.settings import (
    BufferSettings,
    ExplorerSettings,
    LoggerSettings,
    MiscellaneousSettings,
    OptimizerSettings,
    Settings,
)
from pearll.signal_processing.advantage_estimators import generalized_advantage_estimate
from pearll.updaters.actors import BaseActorUpdater, PolicyGradient
from pearll.updaters.critics import BaseCriticUpdater, ValueRegression


def get_default_model(env: Env) -> ActorCritic:
    """
    Returns a default model for the given environment.
    """
    if isinstance(env.action_space, gym.spaces.Discrete):
        action_size = env.action_space.n
    elif isinstance(env.action_space, gym.spaces.Box):
        action_size = env.action_space.shape[0]
    observation_shape = get_space_shape(env.observation_space)

    encoder = IdentityEncoder()
    torso = MLP(layer_sizes=[observation_shape[0], 64, 64], activation_fn=T.nn.Tanh)
    head_actor = CategoricalHead(
        input_shape=64, action_size=action_size, activation_fn=T.nn.Tanh
    )
    head_critic = ValueHead(input_shape=64)

    actor = Actor(encoder=encoder, torso=torso, head=head_actor)
    critic = Critic(encoder=encoder, torso=torso, head=head_critic)

    return ActorCritic(actor=actor, critic=critic)


class A2C(BaseAgent):
    """
    Actor-Critic Algorithm

    :param env: the gym-like environment to be used
    :param model: the neural network model
    :param entropy_coeff: entropy weight
    :param gae_lambda: GAE exponential average coefficient
    :param gae_gamma: trajectory discount factor
    :param value_coeff: value loss weight
    :param actor_updater_class: actor updater class
    :param actor_optimizer_settings: actor optimizer settings
    :param critic_updater_class: critic updater class
    :param critic_optimizer_settings: critic optimizer settings
    :param buffer_class: the buffer class for storing and sampling trajectories
    :param buffer_settings: settings for the buffer
    :param action_explorer_class: the explorer class for random search at beginning of training and
        adding noise to actions
    :param explorer_settings: settings for the action explorer
    :param callbacks: an optional list of callbacks (e.g. if you want to save the model)
    :param callback_settings: settings for callbacks
    :param logger_settings: settings for the logger
    :param misc_settings: settings for miscellaneous parameters
    """

    def __init__(
        self,
        env: Env,
        model: Optional[ActorCritic] = None,
        entropy_coeff: float = 0.01,
        gae_lambda: float = 0.95,
        gae_gamma: float = 0.99,
        value_coeff: float = 0.5,
        actor_updater_class: Type[BaseActorUpdater] = PolicyGradient,
        actor_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        critic_updater_class: Type[BaseCriticUpdater] = ValueRegression,
        critic_optimizer_settings: OptimizerSettings = OptimizerSettings(),
        buffer_class: Type[BaseBuffer] = RolloutBuffer,
        buffer_settings: BufferSettings = BufferSettings(),
        action_explorer_class: Type[BaseExplorer] = BaseExplorer,
        explorer_settings: ExplorerSettings = ExplorerSettings(start_steps=0),
        callbacks: Optional[List[Type[BaseCallback]]] = None,
        callback_settings: Optional[List[Settings]] = None,
        logger_settings: LoggerSettings = LoggerSettings(),
        misc_settings: MiscellaneousSettings = MiscellaneousSettings(),
    ) -> None:
        model = model or get_default_model(env)
        super().__init__(
            env,
            model,
            action_explorer_class=action_explorer_class,
            explorer_settings=explorer_settings,
            buffer_class=buffer_class,
            buffer_settings=buffer_settings,
            logger_settings=logger_settings,
            callbacks=callbacks,
            callback_settings=callback_settings,
            misc_settings=misc_settings,
        )
        self.actor_optimizer_settings = actor_optimizer_settings
        self.critic_optimizer_settings = critic_optimizer_settings
        self.actor_updater = actor_updater_class(
            optimizer_class=actor_optimizer_settings.optimizer_class,
            max_grad=actor_optimizer_settings.max_grad,
        )
        self.critic_updater = critic_updater_class(
            loss_class=critic_optimizer_settings.loss_class,
            optimizer_class=critic_optimizer_settings.optimizer_class,
            max_grad=critic_optimizer_settings.max_grad,
        )
        self.value_coeff = value_coeff
        self.entropy_coeff = entropy_coeff
        self.gae_lambda = gae_lambda
        self.gae_gamma = gae_gamma

    def _fit(self, batch_size: int, actor_epochs: int = 1, critic_epochs: int = 1):
        critic_losses = np.zeros(shape=(critic_epochs))
        actor_losses = np.zeros(shape=(actor_epochs))
        divergences = np.zeros(shape=(actor_epochs))
        entropies = np.zeros(shape=(actor_epochs))

        # Sample rollout and compute advantages/returns
        trajectories = self.buffer.sample(batch_size)
        with T.no_grad():
            old_values = self.model.forward_critics(trajectories.observations)
            next_values = self.model.forward_critics(trajectories.next_observations)
            advantages, returns = generalized_advantage_estimate(
                rewards=trajectories.rewards,
                old_values=old_values,
                new_values=next_values,
                dones=trajectories.dones,
                gae_lambda=self.gae_lambda,
                gamma=self.gae_gamma,
            )

        # Train actor for actor_epochs
        for i in range(actor_epochs):
            actor_log = self.actor_updater(
                model=self.model,
                observations=trajectories.observations,
                actions=trajectories.actions,
                advantages=advantages,
                learning_rate=self.actor_optimizer_settings.learning_rate,
                entropy_coeff=self.entropy_coeff,
            )
            actor_losses[i] = actor_log.loss
            divergences[i] = actor_log.divergence
            entropies[i] = actor_log.entropy

        # Train critic for critic_epochs
        for i in range(critic_epochs):
            critic_log = self.critic_updater(
                self.model,
                trajectories.observations,
                returns,
                learning_rate=self.critic_optimizer_settings.learning_rate,
                loss_coeff=self.value_coeff,
            )
            critic_losses[i] = critic_log.loss

        self.buffer.reset()

        return Log(
            actor_loss=actor_losses.mean(),
            critic_loss=critic_losses.mean(),
            divergence=divergences.sum(),
            entropy=entropies.mean(),
        )



from typing import Union

import numpy as np
from gym import spaces

from pearll.common.type_aliases import Observation
from pearll.explorers.base_explorer import BaseExplorer
from pearll.models.actor_critics import Actor, ActorCritic


class GaussianExplorer(BaseExplorer):
    """
    Add Gaussian noise to the actions

    :param action_space: action space
    :param scale: std of the Gaussian noise
    :param start_steps: the fist n steps to uniformally sample actions
    """

    def __init__(
        self,
        action_space: spaces.Space,
        scale: float = 0.1,
        start_steps: int = 20000,
    ) -> None:
        super().__init__(action_space, start_steps=start_steps)
        self.scale = scale

    def __call__(
        self, model: Union[Actor, ActorCritic], observation: Observation, step: int
    ) -> np.ndarray:
        """
        Get an action for the given observation in training.

        :param model: the model to use
        :param observation: the observation
        :param step: the current training step
        """
        actions = super().__call__(model, observation, step)
        if step >= self.start_steps:
            noises = np.random.normal(loc=0.0, scale=self.scale, size=self.action_shape)
            actions = actions + noises
            actions = np.clip(actions, self.action_range[0], self.action_range[1])
        return actions
